/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PackageAssemblyMojo.java 16:56:43 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.plugin.jbiplugin;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.codehaus.plexus.util.interpolation.ObjectBasedValueSource;
import org.codehaus.plexus.util.interpolation.RegexBasedInterpolator;
import org.objectweb.petals.plugin.jbiplugin.util.PropertiesInterpolationValueSource;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Make a JBI Archive of a project
 * 
 * @goal jbi-package
 * @description Make a JBI archive from a JBI component project
 * @requiresDependencyResolution runtime
 * 
 * @author ddesjardins - eBMWebsourcing
 * @author cdeneux - Capgemini Sud
 */
public class JBIPackageMojo extends JBIAbstractMojo {

    /**
     * Component Jar
     * 
     * @parameter expression="${project.build.directory}/${project.artifactId}-${project.version}.jar"
     * @editable true
     * @required
     * @description Name of the generated component Jar
     */
    protected File componentJar;

    /**
     * Update JBI XML file
     * 
     * @parameter expression="true"
     * @editable true
     * @required
     * @description Boolean to force the update of the JBI XML file to add all
     *              the dependencies of the archive
     */
    protected boolean updateJBIXml;
    
    /**
	 * Sets the mapping pattern for all SU file name included in this service
	 * assembly.
	 * 
	 * @parameter expression="$${artifactId}-$${version}.$${extension}"
	 * @editable true
	 * @required
	 * @description Sets the mapping pattern for all SU file name included in
	 *              this service assembly. Example of usage (caution to the
	 *              usage of "$" and "$$"):
	 *              <serviceUnitFileNameMapping>${artifactId}.${extension}</serviceUnitFileNameMapping>
	 */
    protected String serviceUnitFileNameMapping;

    /**
     * List of entry to exclude (comma separated values)
     * 
     * @parameter expression="foo"
     * @required
     * @editable true
     * @description List of entry to exclude (comma separated values)
     */
    protected String toExcludes;

    /**
     * Set of entries in the zip
     */
    private Set<String> entries = new HashSet<String>();

    /**
     * List of entries to exclude
     */
    private Set<String> entriesToExclude = new HashSet<String>();

    /**
     * Recurse a directory to add its content to a zip file. If updateJBIXml file
     * is true do not add the jbi.xml file to the archive (it will be updated and added later).
     * 
     * @param zipOutputStream
     *            zip file
     * @param directory
     *            directory to recurse
     * @param entryDirectoryName
     *            directory to put the content in the zip
     * @throws Exception
     */
    private void recurseDirectory(ZipOutputStream zipOutputStream,
        File directory, String entryDirectoryName) throws IOException {
        for (File file : directory.listFiles()) {
            if (file.isFile()) {
                if (!updateJBIXml
                    || (updateJBIXml && !file.getName().equals("jbi.xml"))) {
                    
                    if (verbose) {
                        this.getLog().info(
                            "     " + file.getName() + " added to JBI archive");
                    }

                    // Use "/" instead of File.separator to have a directory
                    // level in the ZIP file.
                    ZipEntry zipEntry = new ZipEntry((entryDirectoryName
                        .equals("") ? "" : entryDirectoryName + "/")
                        + file.getName());
                    DataInputStream dis = new DataInputStream(
                        new FileInputStream(file));
                    byte[] content = new byte[dis.available()];
                    dis.readFully(content);
                    zipOutputStream.putNextEntry(zipEntry);
                    zipOutputStream.write(content);
                    zipOutputStream.closeEntry();
                }
            } else {
            	this.getLog().debug("File entry: " + file.getName());
                // Avoid to add the .svn and CVS directories and add directory
                // to zip.
                if (!file.getName().startsWith(".")
                    && !file.getName().toUpperCase().equals("CVS")) {
                    // Use "/" instead of File.separator to have a directory
                    // level in the ZIP file
                    recurseDirectory(zipOutputStream, file, (entryDirectoryName
                        .equals("") ? "" : entryDirectoryName + "/")
                        + file.getName());
                }
            }
        }
    }

    /**
     * Extract the list of jar to exclude from the archive
     * 
     */
    private void extractExcludeValues() {
        // We remove the space
        if (!"foo".equals(toExcludes)) {
            toExcludes = toExcludes.replace(" ", "");
            for (String jar : toExcludes.split(",")) {
                entriesToExclude.add(jar);
            }
        }
    }

    /**
     * Execute the plugin
     */
    public void execute() throws MojoExecutionException, MojoFailureException {
            project.getArtifact().setFile(new File(outputDirectory
                + File.separator + jbiName + ".zip"));
        
        // Extract the exclude values
        extractExcludeValues();

        if (outputDirectory.exists()) {
            if (project.getArtifact().getFile().exists()) {
                project.getArtifact().getFile().delete();
            }
        } else {
            outputDirectory.mkdirs();
        }

            if (jbiDirectory.exists()) {
            if (PACKAGING_COMPONENT.equals(project.getPackaging())) {
                packageComponent();
            } else if (PACKAGING_SU.equals(project.getPackaging())) {
                packageServiceUnit();
            } else if (PACKAGING_SA.equals(project.getPackaging())) {
                packageServiceAssembly();
            } else if (PACKAGING_SL.equals(project.getPackaging())) {
                packageSharedLibrary();
            }
        } else {
            this.getLog().error("JBI directory [" + jbiDirectory + "] does not exist.");
            throw new MojoFailureException("JBI directory [" + jbiDirectory + "] does not exist.");
        }
                }

    /**
     * Create a service-unit package.
     */
    private void packageComponent() throws MojoExecutionException {
        try {
            this.getLog().info("Start building JBI component archive "
                + project.getArtifact().getFile().getAbsolutePath());

            // Add the jbi files
            ZipOutputStream zipOutputStream = new ZipOutputStream(
            new FileOutputStream(project.getArtifact().getFile()));
            recurseDirectory(zipOutputStream, jbiDirectory, "");
            zipOutputStream.flush();

            // Add the component jar
            addComponentResources(zipOutputStream);

            // Add the component dependencies
            addDependencies(zipOutputStream);

            try {
                if (updateJBIXml) {
                    // Update the JBI XML file
                    updateComponentJBIXmlFile(zipOutputStream);
                }
            }
            catch (FileNotFoundException e) {
                if (verbose) {
                    this.getLog().error("Fail to add jbi.xml file");
                }
                throw e;
            }

            zipOutputStream.flush();
            zipOutputStream.close();
            
            this.getLog().info("JBI component archive building done.");
        } catch (IOException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }
    
    
    /**
     * Create a service-unit package.
     */
    private void packageServiceUnit() throws MojoExecutionException {
        try {
            this.getLog().info(
                "Start building JBI service-unit archive "
                    + project.getArtifact().getFile().getAbsolutePath());
            
            // JBI Archive creation
            ZipOutputStream zipOutputStream = new ZipOutputStream(
                new FileOutputStream(project.getArtifact().getFile()));
            
            // Add the jbi files
            recurseDirectory(zipOutputStream, jbiDirectory, "");
            zipOutputStream.flush();
            
            // Add directly the jbi.xml descriptor
            try {
				DataInputStream jbiDescriptorDis = new DataInputStream(
						new FileInputStream(jbiDirectory + File.separator + "jbi.xml"));
				byte[] jbiDescriptorContent = new byte[jbiDescriptorDis
						.available()];
				jbiDescriptorDis.readFully(jbiDescriptorContent);

				ZipEntry jbiDescriptorZipEntry = new ZipEntry("META-INF/jbi.xml");
				zipOutputStream.putNextEntry(jbiDescriptorZipEntry);
				zipOutputStream.write(jbiDescriptorContent);
				zipOutputStream.closeEntry();
			} catch (FileNotFoundException e) {
                if (verbose) {
					this.getLog().error("Fail to add jbi.xml file");
                }
                throw e;
            }

            // Add the service unit resources
            addComponentResources(zipOutputStream);
            
            // Add the service unit dependencies
            addDependencies(zipOutputStream);
            
            zipOutputStream.close();
            
            this.getLog().info("JBI Service unit archive building done.");

        } catch (IOException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }
    
    /**
     * Create a service-assembly package.
     */
    private void packageServiceAssembly() throws MojoExecutionException {
        try {
            this.getLog().info(
                "Start building JBI service-assembly archive "
                    + project.getArtifact().getFile().getAbsolutePath());

            // JBI Archive creation
            ZipOutputStream zipOutputStream = new ZipOutputStream(
                new FileOutputStream(project.getArtifact().getFile()));
            
            // Add the jbi.xml file
            try {
				DataInputStream jbiDescriptorDis = new DataInputStream(
						new FileInputStream(jbiDirectory + File.separator
								+ "jbi.xml"));
				byte[] jbiDescriptorContent = new byte[jbiDescriptorDis
						.available()];
				jbiDescriptorDis.readFully(jbiDescriptorContent);

				ZipEntry jbiDescriptorZipEntry = new ZipEntry(
						"META-INF/jbi.xml");
				zipOutputStream.putNextEntry(jbiDescriptorZipEntry);
				zipOutputStream.write(jbiDescriptorContent);
				zipOutputStream.closeEntry();
            } catch (FileNotFoundException e) {
				if (verbose) {
					this.getLog().error("Fail to add jbi.xml file");
				}
                throw e;
			}

            // Add the service unit dependencies files
            Set<Artifact> dependencies = project.getDependencyArtifacts();
            for (Artifact dependency : dependencies) {

                if (PACKAGING_SU.equals(dependency.getType())) {
                    String zipDependencyFileName = evaluateFileNameMapping(this.serviceUnitFileNameMapping, dependency);

                    DataInputStream dis = new DataInputStream(new FileInputStream(dependency.getFile()));
                    byte[] content = new byte[dis.available()];
                    dis.readFully(content);
    
                    ZipEntry zipEntry = new ZipEntry(zipDependencyFileName);
                    zipOutputStream.putNextEntry(zipEntry);
                    zipOutputStream.write(content);
                    zipOutputStream.closeEntry();
                    if (verbose) {
                        this.getLog().info("     "
                            + dependency.getFile().getName() + " added to JBI Service Assembly archive (as " + zipDependencyFileName + ")");
                    }
                }
            }
            
            zipOutputStream.close();

            this.getLog().info("JBI Service Assembly archive building done.");

        } catch (IOException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }

    /**
     * Create a shared-library package.
     */
    private void packageSharedLibrary() throws MojoExecutionException{
        try {
            this.getLog().info(
                "Start building JBI shared-library archive "
                    + project.getArtifact().getFile().getAbsolutePath());

            // JBI Archive creation
            ZipOutputStream zipOutputStream = new ZipOutputStream(
                new FileOutputStream(project.getArtifact().getFile()));
            
            // Add the jbi files
            recurseDirectory(zipOutputStream, jbiDirectory, "");
            zipOutputStream.flush();

            // Add the shared-library component jar
            addComponentResources(zipOutputStream);

            // Add the shared library dependencies
            addDependencies(zipOutputStream);
            
            try {
                if (updateJBIXml) {
                    // Update the JBI XML file
                    updateSharedLibraryJBIXmlFile(zipOutputStream);
                }
            }
            catch (FileNotFoundException e) {
                if (verbose) {
                    this.getLog().error("Fail to add jbi.xml file");
                }
                throw e;
            }
            
            zipOutputStream.close();

            this.getLog().info("JBI Shared library archive building done.");

        } catch (IOException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }
    
    
    /**
     * Add the Java resources (classes and resources) in the JBI archive.
     * @param zipOutputStream The JBI archive
     */
    private void addComponentResources(ZipOutputStream zipOutputStream) throws IOException {
        if (componentJar.exists()) {
            zipFile(zipOutputStream, componentJar);
        }
    }

    /**
     * Add dependencies in the JBI archive.
     * 
     * @param zipOutputStream
     *            The JBI archive
     */
    private void addDependencies(ZipOutputStream zipOutputStream)
        throws IOException {

        Set<Artifact> artifacts = project.getArtifacts();
        for (Artifact artifact : artifacts) {
            
            String artifactName = artifact.getArtifactId() + "-"
                + artifact.getVersion() + ".jar";

            boolean dependencyToInclude = false;
            
            if (!artifact.isOptional()
                && (artifact.getScope() == null || (artifact.getScope() != null
                    && !artifact.getScope().equals("test") && !artifact
                    .getScope().equals("provided")))) {
                
                dependencyToInclude = true;
                
            }
            if (dependencyToInclude && !entriesToExclude.contains(artifactName)) {
                zipFile(zipOutputStream, artifact.getFile());
            }
        }
    }

    /**
     * Clone a set of String
     * 
     * @param setToClone
     * @return
     */
    private Set<String> cloneSet(Set<String> setToClone) {
        Set<String> outSet = new HashSet<String>();
        for (String string : setToClone) {
            outSet.add(string);
        }
        return outSet;
    }

    /**
     * Update the JBI xml file to added all the dependencies of a component JBI
     * archive.
     * 
     * @param zipOutputStream
     * @throws FileNotFoundException
     */
    private void updateComponentJBIXmlFile(ZipOutputStream zipOutputStream)
        throws FileNotFoundException {
        InputStream inputStream = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory
                .newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            Document document = builder.parse(new File(jbiDirectory
                + File.separator + "jbi.xml"));
            Element element = document.getDocumentElement();

            // Check the component-class-path

            Element compoClassPath = (Element) element.getElementsByTagName(
                "component-class-path").item(0);
            this.updateClasspathBlock(compoClassPath, document);
            // add the main component-jar in the component-class-path
            Element mainElem = document.createElement("path-element");
            mainElem.setTextContent(componentJar.getName());
            compoClassPath.appendChild(mainElem);
            
            // Check the bootstrap-class-path
            Element bootClassPath = (Element) element.getElementsByTagName(
                "bootstrap-class-path").item(0);
            this.updateClasspathBlock(bootClassPath, document);
            // add the main component-jar in the component-class-path
            mainElem = document.createElement("path-element");
            mainElem.setTextContent(componentJar.getName());
            compoClassPath.appendChild(mainElem);
            
            document.normalize();
            Source source = new DOMSource(document);
            StringWriter out = new StringWriter();
            Result resultStream = new StreamResult(out);
            TransformerFactory tFactory = TransformerFactory.newInstance();
            Transformer transformer;
            transformer = tFactory.newTransformer();
            transformer.transform(source, resultStream);

            inputStream = new ByteArrayInputStream(out.toString().getBytes());

        } catch (Exception e) {
            if (verbose) {
                System.out
                    .println("Problem while updating jbi.xml file. We add it like it is");
            }
            inputStream = new FileInputStream(new File(jbiDirectory
                + File.separator + "jbi.xml"));
        }

        // Add the jbi.xml entry
        try {
            ZipEntry zipEntry = new ZipEntry("META-INF/jbi.xml");
            DataInputStream dis = new DataInputStream(inputStream);
            byte[] content = new byte[dis.available()];
            dis.readFully(content);
            zipOutputStream.putNextEntry(zipEntry);
            zipOutputStream.write(content);
            zipOutputStream.closeEntry();
        } catch (Exception e) {
            if (verbose) {
                System.out.println("Fail to add jbi.xml file");
            }
        }
    }

    /**
     * Update the JBI xml file to added all the dependencies of a shared library
     * JBI archive.
     * 
     * @param zipOutputStream
     * @throws FileNotFoundException
     */
    private void updateSharedLibraryJBIXmlFile(ZipOutputStream zipOutputStream)
        throws FileNotFoundException {
        InputStream inputStream = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory
                .newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            Document document = builder.parse(new File(jbiDirectory
                + File.separator + "jbi.xml"));
            Element element = document.getDocumentElement();

            // Check the shared-library-class-path
            Element sharedLibraryClassPath = (Element) element.getElementsByTagName(
                "shared-library-class-path").item(0);
            this.updateClasspathBlock(sharedLibraryClassPath, document);
            // add the main component-jar in the shared-library-class-path
            Element mainElem = document.createElement("path-element");
            mainElem.setTextContent(componentJar.getName());
            sharedLibraryClassPath.appendChild(mainElem);

            document.normalize();
            Source source = new DOMSource(document);
            StringWriter out = new StringWriter();
            Result resultStream = new StreamResult(out);
            TransformerFactory tFactory = TransformerFactory.newInstance();
            Transformer transformer;
            transformer = tFactory.newTransformer();
            transformer.transform(source, resultStream);

            inputStream = new ByteArrayInputStream(out.toString().getBytes());

        } catch (Exception e) {
            if (verbose) {
                System.out
                    .println("Problem while updating jbi.xml file. We add it like it is");
            }
            inputStream = new FileInputStream(new File(jbiDirectory
                + File.separator + "jbi.xml"));
        }

        // Add the jbi.xml entry
        try {
            ZipEntry zipEntry = new ZipEntry("META-INF/jbi.xml");
            DataInputStream dis = new DataInputStream(inputStream);
            byte[] content = new byte[dis.available()];
            dis.readFully(content);
            zipOutputStream.putNextEntry(zipEntry);
            zipOutputStream.write(content);
            zipOutputStream.closeEntry();
        } catch (Exception e) {
            if (verbose) {
                System.out.println("Fail to add jbi.xml file");
            }
        }
    }
    
    private void updateClasspathBlock(Element classpathElement, Document document) {
        NodeList bootElems = classpathElement.getElementsByTagName("path-element");
        Set<String> elemsToAdd = cloneSet(entries);
        for (int i = 0; i < bootElems.getLength(); i++) {
            Element elem = (Element) bootElems.item(i);
            if (elemsToAdd.contains(elem.getTextContent())) {
                elemsToAdd.remove(elem.getTextContent());
            }
        }
        if (elemsToAdd.size() > 0) {
            // We need to add the missing elements
            for (String jar : elemsToAdd) {
                Element elem = document.createElement("path-element");
                elem.setTextContent(jar);
                classpathElement.appendChild(elem);
            }
        }
    }

    /**
     * Add a file to a zip stream
     * 
     * @param zipOutputStream
     * @param file
     */
    private void zipFile(ZipOutputStream zipOutputStream, File file) {

        try {
            if (!entries.contains(file.getName())) {
                DataInputStream dis = new DataInputStream(new FileInputStream(
                    file));
                byte[] content = new byte[dis.available()];
                dis.readFully(content);

                ZipEntry zipEntry = new ZipEntry(file.getName());
                zipOutputStream.putNextEntry(zipEntry);
                zipOutputStream.write(content);
                zipOutputStream.closeEntry();

                entries.add(file.getName());

                if (verbose) {
                    this.getLog().info(
                        "     " + file.getName() + " added to JBI archive");
                }
            }
        } catch (IOException e) {
            entriesToExclude.add(file.getName());
            if (verbose) {
                this.getLog().error(
                    "Problem while adding " + file.getName()
                        + " to JBI archive");
            }
        }
    }
    
    /**
     * Evaluates Filename Mapping (source code copied from maven-assembly-plugin)
     * 
     * @param expression
     * @param artifact
     * @return expression
     * @throws AssemblyFormattingException 
     * @throws org.apache.maven.plugin.MojoExecutionException
     */
    private String evaluateFileNameMapping( String expression, Artifact artifact )
    {
    	String value = expression.replaceAll("\\$\\{dollarSign\\}", "\\$");

        // FIXME: This is BAD! Accessors SHOULD NOT change the behavior of the object.
        artifact.isSnapshot();

        RegexBasedInterpolator interpolator = new RegexBasedInterpolator();
        
        interpolator.addValueSource( new ObjectBasedValueSource( artifact ) );
        interpolator.addValueSource( new ObjectBasedValueSource( artifact.getArtifactHandler() ) );
        
        Properties classifierMask = new Properties();
        classifierMask.setProperty( "classifier", "" );
        
        interpolator.addValueSource( new PropertiesInterpolationValueSource( classifierMask ) );
        
        value = interpolator.interpolate( value, "__artifact" );

        return value;
    }
}
